from functools import partial
from threading import Thread
from pathlib import Path
from queue import Empty
from typing import Tuple

from ipcqueue.posixmq import Queue

from cobra_py import rl
from cobra_py.raylib import ffi

sw = 800
sh = 450

window = rl.init_window(sw, sh, b"Cobra Py!")

font_path = str(
    Path(rl.__file__).parent / "resources" / "fonts" / "Kepler-452b.ttf"
).encode("utf-8")
font = rl.load_font_ex(font_path, 24, ffi.NULL, 0)
text_size = rl.measure_text_ex(font, b"X", font.baseSize, 0)
sw = int(80 * text_size.x)
sh = int(25 * text_size.y)
rl.set_window_size(sw, sh)
rl.set_target_fps(60)

rl.set_trace_log_level(rl.LOG_NONE)


class Screen:
    def __init__(self):
        self.command_queue = Queue('/foo')
        self._screen = []
        for y in range(25):  # rows
            self._screen.append([])
            for x in range(80):  # columns
                self._screen[-1].append(b" ")
        self._buffer = rl.gen_image_color(sw, sh, (rl.RAYWHITE))
        self._buffer_texture = None
        self._update_buffer_texture()

    def _update_buffer_texture(self):
        if self._buffer_texture is not None:
            rl.unload_texture(self._buffer_texture)
        self._buffer_texture = rl.load_texture_from_image(self._buffer)

    def event_loop(self):
        rl.begin_drawing()
        # Draw graphics buffer
        rl.draw_texture(self._buffer_texture, 0, 0, rl.WHITE)
        # Draw text mode screen
        for y, row in enumerate(self._screen):
            text = b"".join(row)
            rl.draw_text_ex(
                font,
                text,
                (0, y * text_size.y),
                font.baseSize,
                0,
                rl.RED,
            )
        rl.draw_fps(10, 10)
        rl.end_drawing()

    def process_commands(self):
        try:
            while True:
                command, a, kw = self.command_queue.get(False)
                getattr(self, command)(*a, **kw)
                # self.command_queue.task_done()
        except Empty:
            pass

    def print_at(self, x: int = 0, y: int = 0, text: str = ""):
        for i, c in enumerate(text):
            self._screen[y][x + i % 80] = c.encode("utf-8")

    def circle(self, x: int, y: int, radius: int, color: Tuple[int, int, int, int]):
        rl.image_draw_circle(ffi.addressof(self._buffer), x, y, radius, color)
        self._update_buffer_texture()


screen = Screen()


def command(cmdname, *a, **kw):
    screen.command_queue.put((cmdname, a, kw))


print_at = partial(command, "print_at")
circle = partial(command, "circle")


def prompt():
    while True:
        cmd = input("> ")
        if cmd:
            eval(cmd)


x = Thread(target=prompt)
x.start()

screen.circle(100, 100, 50, rl.RED)
screen.print_at(10, 10, "FOOBAR")

while not rl.window_should_close():
    screen.process_commands()
    screen.event_loop()



"""A wrapper to the raylib CFFI binding so that names are snake_case instead of CamelCase"""

import inspect
import re
from types import BuiltinFunctionType

import cobra_py.raylib.lib as rl


def camel_to_snake(name):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


# Doesn't work with CFFI functions (yet!) but hey, it's an idea
def wrap_bytes(f):
    try:
        sig = inspect.signature(f)
    except ValueError:
        return f
    byte_args = [
        i for i, p in enumerate(sig.parameters) if sig.parameters[p].annotation == bytes
    ]

    def g(*a):
        _a = list(a)
        for i in byte_args:
            _a[i] = _a[i].encode("utf-8")
        f(*a)

    return g


for name, value in inspect.getmembers(rl):
    if isinstance(value, BuiltinFunctionType):
        locals()[camel_to_snake(name)] = wrap_bytes(value)
    locals()[name] = value



import os
import pathlib

import cffi
from invoke import task


@task
def checkout_raylib(c):
    if os.path.isdir("raylib"):
        return
    c.run("git clone https://github.com/raysan5/raylib.git")
    with c.cd("raylib"):
        c.run("git checkout 3.0.0")
        c.run("patch -p1 < ../raylib.patch")


@task
def build_raylib(c):
    with c.cd("raylib/src"):
        c.run("make clean")
        c.run("make PLATFORM=PLATFORM_DESKTOP")


@task
def check(c):
    print("Formatting")
    c.run("isort .")
    c.run("black .")


@task(build_raylib)
def bind_raylib(c):
    print("Building raylib CFFI")
    ffi = cffi.FFI()
    includes = pathlib.Path().absolute() / "include"
    lib_dir = pathlib.Path().absolute() / "raylib" / "src"
    h_file = includes / "_raylib.h"
    with open(h_file) as inf:
        ffi.cdef(inf.read())
    ffi.set_source(
        "cobra_py.raylib",
        """
            #define RAYGUI_IMPLEMENTATION
            #define RAYGUI_SUPPORT_ICONS
            #include "raylib.h"
            #include "raygui.h"
        """,
        libraries=["raylib", "X11"],
        library_dirs=[lib_dir.as_posix()],
        extra_link_args=["-Wl,-rpath,."],
        include_dirs=[includes, lib_dir],
    )
    ffi.compile()



from cobra_py import rl

sw = 800
sh = 450

# Same as the raylib example, just using snake_case instead of CamelCase for functions
rl.init_window(sw, sh, b"Example")
rl.set_target_fps(60)
while not rl.window_should_close():
    rl.begin_drawing()
    rl.clear_background(rl.RAYWHITE)
    rl.draw_text(
        b"Congrats! You created your first window!",
        190,
        200,
        20,
        rl.RED,
    )
    rl.end_drawing()
# Actually you can still use CamelCase, be free
rl.CloseWindow()



from cobra_py import rl

sw = 800
sh = 450

rl.init_window(sw, sh, b"CheckBoxes")
rl.set_target_fps(60)
cb = 1

while not rl.window_should_close():
    rl.begin_drawing()
    rl.clear_background(rl.RAYWHITE)
    cb = rl.gui_check_box((300, 300, 20, 20), b"Foo", cb)
    rl.draw_fps(10, 10)
    rl.end_drawing()
rl.close_window()



from pathlib import Path

from cobra_py import rl
from cobra_py.raylib import ffi

sw = 800
sh = 450

rl.init_window(sw, sh, b"Raylib Fonts")

font_path = str(
    Path(rl.__file__).parent / "resources" / "fonts" / "Kepler-452b.ttf"
).encode("utf-8")
kepler_font = rl.LoadFontEx(font_path, 48, ffi.NULL, 0)

message = b"This is the Kepler Font"
text_size = rl.MeasureTextEx(kepler_font, message, kepler_font.baseSize, 0)

while not rl.window_should_close():
    rl.begin_drawing()
    rl.clear_background(rl.RAYWHITE)
    rl.draw_text_ex(
        kepler_font,
        message,
        ((sw - text_size.x) / 2, (sh - text_size.y) / 2),
        kepler_font.baseSize,
        0,
        rl.RED,
    )
    rl.end_drawing()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cobra_py']

package_data = \
{'': ['*'], 'cobra_py': ['resources/fonts/*']}

install_requires = \
['glooey>=0.3.4,<0.4.0',
 'ipcqueue>=0.9.6,<0.10.0',
 'pillow>=7.2.0,<8.0.0',
 'pyglet>=1.5.7,<2.0.0',
 'pyte>=0.8.0,<0.9.0']

setup_kwargs = {
    'name': 'cobra-py',
    'version': '0.0.2',
    'description': 'An 80s style Python runtime and development environment',
    'long_description': None,
    'author': 'Roberto Alsina',
    'author_email': 'roberto.alsina@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python

import os
import pty
import select
from pathlib import Path

import pyte

from cobra_py import rl
from cobra_py.raylib import ffi

sw = 800
sh = 450


rl.init_window(sw, sh, b"CobraPy Terminal!")
font_path = str(Path(rl.__file__).parent / "resources" / "fonts" / "monoid.ttf").encode(
    "utf-8"
)
font = rl.load_font_ex(font_path, 24, ffi.NULL, 0)
text_size = rl.measure_text_ex(font, b"X", font.baseSize, 0)
sw = int(80 * text_size.x)
sh = int(25 * text_size.y)
rl.set_window_size(sw, sh)
rl.set_target_fps(24)


term = pyte.HistoryScreen(80, 25)
stream = pyte.ByteStream(term)

p_pid, master_fd = pty.fork()
if p_pid == 0:  # Child process
    os.execvpe(
        "bash",
        ["bash"],
        env=dict(TERM="xterm", COLUMNS="80", LINES="25", LC_ALL="en_US.UTF-8"),
    )

p_out = os.fdopen(master_fd, "w+b", 0)


def esc(d: bytes):
    return b"\x1b" + d


keys = {
    36: b"\n",  # Enter
    23: b"\t",  # Tab
    66: b"\x1b",  # ESC
    111: esc(b"[A"),  # Cursor up
    116: esc(b"[B"),  # Cursor down
    114: esc(b"[C"),  # Cursor right
    113: esc(b"[D"),  # Cursor left
    22: b"\b",  # Backspace
    119: esc(b"[3~"),  # Delete
    67: esc(b"OP"),  # F1
    68: esc(b"OQ"),  # F2
    69: esc(b"OR"),  # F3
    70: esc(b"OS"),  # F4
    71: esc(b"[15~"),  # F5
    72: esc(b"[17~"),  # F6
    73: esc(b"[18~"),  # F7
    74: esc(b"[19~"),  # F8
    75: esc(b"[20~"),  # F9
    76: esc(b"[21~"),  # F10
    95: esc(b"[23~"),  # F11
    96: esc(b"[24~"),  # F12
}

ctrl_keys = {40: b"\x04", 54: b"\x03", 27: b"\x12"}  # D  # C  # R

colors = {
    "black": rl.BLACK,
    "red": rl.RED,
    "green": rl.GREEN,
    "brown": rl.BROWN,
    "blue": rl.BLUE,
    "magenta": rl.MAGENTA,
    "cyan": (0, 255, 255, 255),
    "white": rl.WHITE,
}

ctrl = False


@ffi.callback("void(int,int,int,int)")
def key_event(key, scancode, action, mods):
    global ctrl
    print("in-scancode=>", key, scancode, action, ctrl, mods)
    if mods == 0:  # Key release
        if action == 37:  # ctrl
            print("ctrl-off")
            ctrl = False
        print("out-scancode=>", key, scancode, action, ctrl, mods)
        return

    # Key press or repeat
    if action == 37:
        print("ctrl-on")
        ctrl = True
    if ctrl and mods == 1:  # ctrl-key doesn't repeat
        if data := ctrl_keys.get(action):
            print("--->", repr(data))
            p_out.write(data)
    elif data := keys.get(action):
        print("--->", repr(data))
        p_out.write(data)
    print("out-scancode=>", key, scancode, action, ctrl, mods)


rl.set_key_callback(key_event)


while not rl.window_should_close():
    # FIXME: can probably handle it all in key_event
    key = rl.get_key_pressed()
    while key > 0:
        k = chr(key).encode("utf-8")
        p_out.write(k)
        key = rl.get_key_pressed()

    rl.begin_drawing()
    rl.clear_background(rl.DARKBLUE)

    ready, _, _ = select.select([p_out], [], [], 0)
    if ready:
        data = p_out.read(1024)
        if data:
            stream.feed(data)

    for y, line in enumerate(term.display):
        line = term.buffer[y]
        for x in range(term.columns):  # Can't enumerate, it's sparse
            char = line[x]

            fg = colors.get(char.fg, rl.WHITE)
            bg = colors.get(char.bg, rl.BLACK)
            if char.reverse:
                fg, bg = bg, fg

            rl.draw_rectangle(
                int(x * text_size.x),
                int(y * text_size.y),
                int(text_size.x),
                int(text_size.y),
                bg,
            )
            rl.draw_text_ex(
                font,
                char.data.encode("utf-8"),
                (x * text_size.x, y * text_size.y),
                font.baseSize,
                0,
                fg,
            )
    term.dirty.clear()

    # draw cursor
    rl.draw_rectangle(
        int(term.cursor.x * text_size.x),
        int(term.cursor.y * text_size.y),
        int(text_size.x),
        int(text_size.y),
        (100, 108, 100, 100),
    )
    rl.draw_fps(850, 10)
    rl.end_drawing()
